import unittest

import numpy as np
import pymc as pm

from quad5.quadratic_approximation import QuadraticApproximation


class TestQuap(unittest.TestCase):
    # Example originates from Bayesian Data Analyses, 3rd Edition
    # By Andrew Gelman, John Carlin, Hal Stern, David Dunson, Aki Vehtari, and Donald Rubin.
    # See section. 4.1
    def test_analytical_solution_against_quadratic_approx(
        self,
    ):
        y = np.array([2642, 3503, 4358], dtype=np.float64)
        n = y.size

        with pm.Model() as m:
            logsigma = pm.Uniform("logsigma", 1, 100)
            mu = pm.Uniform("mu", 0, 10000)
            _ = pm.Normal("y", mu=mu, sigma=pm.math.exp(logsigma), observed=y)
            custom_step = QuadraticApproximation([mu, logsigma], m)
            trace = pm.sample(draws=1000, chains=4, tune=100, step=custom_step)

        bda_map = [y.mean(), np.log(y.std())]
        bda_cov = np.array([[y.var() / n, 0], [0, 1 / (2 * n)]])

        assert np.allclose(custom_step.mode, bda_map)
        assert np.allclose(custom_step.covariance, bda_cov, atol=1e-3)

        print(trace.posterior.mean()["mu"])


if __name__ == "__main__":
    unittest.main()



import numpy as np
import pymc as pm
import scipy.stats as st
from pymc.step_methods.arraystep import ArrayStep
from pymc.util import get_value_vars_from_user_vars


class QuadraticApproximation(ArrayStep):
    def __init__(self, vars, model, start=None, **kwargs):
        self.model = model
        self.vars = vars
        self.varnames = [var.name for var in vars]

        self.mode, self.covariance = self._compute_mode_and_covariance(start)

        vars = get_value_vars_from_user_vars(vars, model)

        super().__init__(vars, [self._logp_fn], **kwargs)

    def _point_to_array(self, point):
        return np.array([point[varname] for varname in self.varnames])

    def _array_to_point(self, array):
        return {varname: val for varname, val in zip(self.varnames, array)}

    def _logp_fn(self, x):
        point = self._array_to_point(x)
        return self.model.logp(point)

    def _compute_mode_and_covariance(self, start=None):

        map = pm.find_MAP(vars=self.vars, start=start)

        m = pm.modelcontext(None)

        for var in self.vars:
            if m.rvs_to_transforms[var] is not None:
                m.rvs_to_transforms[var] = None
                var_value = m.rvs_to_values[var]
                var_value.name = var.name

        H = pm.find_hessian(map, vars=self.vars)
        cov = np.linalg.inv(H)
        mean = np.concatenate([np.atleast_1d(map[v.name]) for v in self.vars])

        return mean, cov

    def astep(self, q0, logp):
        sample = np.random.multivariate_normal(self.mode, self.covariance)
        return sample, []

    def posterior(self):
        return st.multivariate_normal(mean=self.mean, cov=self.covariance)



import codecs
import os

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="quad5",
    version="0.1.7",
    description="""Quadratic Approximation custom step sampler for PYMC.""",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Carsten Jørgensen",
    packages=find_packages(include=["quad5"]),
    keywords=["Bayesian Statistics", "Probabalistic Programming Language", "Python"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=["PYMC==5.15.0"],
    license="MIT",
    url="https://github.com/carsten-j/quad5",
)
